package igblan.life;

public class Box
{
	public static final Box empty = new Box(0, 0, 0, 0);
	
	public final Cell leftTop, rightBottom;
	
	public Box(Cell leftTop, Cell rightBottom)
	{
		this.leftTop = leftTop;
		this.rightBottom = rightBottom;
	}
	
	public Box(int left, int top, int right, int bottom)
	{
		this.leftTop = new Cell(left, top);
		this.rightBottom = new Cell(right, bottom);
	}
	
	public boolean equals(Box box)
	{
		return leftTop.equals(box.leftTop) &&
			   rightBottom.equals(box.rightBottom);
	}
	
	public Box union(Box box)
	{
		return new Box (leftTop.min(box.leftTop),
						rightBottom.max(box.rightBottom));
	}
	
	public int left() {return leftTop.x;}
	public int top() {return leftTop.y;}
	public int right() {return rightBottom.x;}
	public int bottom() {return rightBottom.y;}
	public int width() {return rightBottom.x - leftTop.x;}
	public int height() {return rightBottom.y - leftTop.y;}
	
	public boolean isEmpty()
		{return leftTop.isZero() && rightBottom.isZero();}
	
	public Box plus(int x, int y)
		{return new Box(leftTop.plus(x, y), rightBottom.plus(x, y));}
	
	public Box plus(Cell p) {return plus(p.x, p.y);}
	
	public Cell extent() {return rightBottom.minus(leftTop);}

	@Override
	public String toString()
	{
		return "[" + left() + "," + top() + "," +
			   right() + "," + bottom() + '}';
	}

	public boolean contains(int x, int y)
		{return x >= left() && x < right() && y >= top() && y < bottom();}
	
	public boolean intersects(Box box)
	{
		return
			Math.max(left(), box.left()) <= Math.min(right(), box.right()) &&
			Math.max(top(), box.top()) <= Math.min(bottom(), box.bottom());
	}
	
	public Cell center()
		{return new Cell((left() + right()) / 2, (top() + bottom()) / 2);}
	
	public Box translate(int x, int y)
		{return new Box(leftTop.plus(x, y), rightBottom.plus(x, y));}

	public Box translate(Cell p) {return translate(p.x, p.y);}
}
