package igblan.seedsOfDestruction;

import igblan.life.Box;
import igblan.life.Cell;
import igblan.life.LifePattern;
import igblan.seedsOfDestruction.ui.Frame;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.Serializable;
import javax.swing.tree.DefaultMutableTreeNode;

/**
 *
 * @author Paul Chapman
 */
public class PuzzleNode extends DefaultMutableTreeNode implements Serializable
{
	private static final long serialVersionUID = 1L;

	private String name;
	private String description;
	private LifePattern pattern;
	private LifePattern smear;

	public PuzzleNode(String name, String description,
					  LifePattern pattern, LifePattern smear)
	{
		super();
		this.name = name;
		this.description = description;
		this.pattern = pattern;
		this.smear = smear.or(pattern);
	}
	
	public PuzzleNode(String name, LifePattern pattern)
	{
		this(name, "", pattern, pattern.getEmpty());
	}
	
	public PuzzleNode(String name, String description, String lifeHistoryRLE)
	{
		super();
		
		this.name = name;
		this.description = description;
		
		String rle = lifeHistoryRLE.replace('.', 'b').replace('A', 'o')
								   .replace('D', 'B');
		
		String patternRLE = rle.replace('B', 'b');
		String smearRLE = rle.replace('B', 'o');

		pattern = Frame.emptyPattern.newPatternFromRLE(patternRLE);
		smear = Frame.emptyPattern.newPatternFromRLE(smearRLE);
		
		Box box = smear.getBoundingBox();
		Cell offset = box.center().negated();
		
		pattern = pattern.translate(offset);
		smear = smear.translate(offset);
	}
	
	public PuzzleNode(String name)
	{
		this(name, "!", "");
	}
	
	public String getName() {return name;}
	public String getDescription() {return description;}
	public LifePattern getPattern() {return pattern;}
	public LifePattern getSmear() {return smear;}
	public boolean isSeedNode() {return getLevel() > 1;}
	
	@Override
	public String toString()
	{
		return isSeedNode() ? "<html><i>" + name + "</i></html>" : name;
	}
	
	public PuzzleNode deepCopy()
	{
		PuzzleNode result = new PuzzleNode(name, description, pattern, smear);
		int count = getChildCount();
		
		for (int i = 0; i < count; i++)
			result.insert(((PuzzleNode) getChildAt(i)).deepCopy(), i);
		
		return result;
	}
	
	public PuzzleNode deepCopyWith(String newDescription,
								   LifePattern newPattern, LifePattern newSmear)
	{
		PuzzleNode result =
			new PuzzleNode(name, newDescription, newPattern, newSmear);
		int count = getChildCount();
		
		for (int i = 0; i < count; i++)
			result.insert(((PuzzleNode) getChildAt(i)).deepCopy(), i);
		
		return result;
	}
	
	private void writeObject(java.io.ObjectOutputStream out) throws IOException
	{
		out.writeObject(name);
		out.writeObject(description);
		out.writeObject(new SerializablePattern(pattern));
		out.writeObject(new SerializablePattern(smear));
	}
	
	private void readObject(ObjectInputStream in)
		throws IOException, ClassNotFoundException
	{
		name = (String) in.readObject();
		
		Object object = in.readObject();
		
		if (object.getClass() == String.class)
		{
			description = (String) object;
			object = in.readObject();
		}
		else description = "";
		
		pattern = ((SerializablePattern) object).toPattern();
		smear = ((SerializablePattern) in.readObject()).toPattern();
	}
}
