package igblan.seedsOfDestruction;

import igblan.life.Box;
import igblan.life.Cell;
import igblan.life.LifePattern;
import igblan.seedsOfDestruction.ui.Frame;
import java.awt.Toolkit;
import java.awt.datatransfer.Clipboard;
import java.awt.datatransfer.StringSelection;
import javax.swing.tree.DefaultTreeModel;
import javax.swing.tree.TreeNode;
import javax.swing.tree.TreePath;

public final class PuzzleTree extends DefaultTreeModel
{
	private static int count = 0;
	
	private static final int SMEAR_INTERVAL = 4;

	public static final int INIT_GENS = 1024;
	public static final int MIN_GENS = 8;
	public static final int MAX_GENS = 4096;
	public static final int START_GEN_INC = 8;

	private int SCORE_BASE = 2;
	
	private static final int[] log2 = {0, 0, 1, 0, 2, 0, 0, 0, 3};
	
	private static int mod(int x, int y)
	{
		return (x + Math.max((-x / y + 1) * y , 0)) % y;
	}
	
	private static final String[][] seedRLEs =
	{
		{"block", "B4", "<html>Block</html>",
			"2o$2o"},
		{"tub", "T3", "<html>Tub</html>",
			"bo$obo$bo"},
		{"pond", "P8", "<html>Pond</html>",
			"b2o$o2bo$o2bo$b2o"},
		{"beehive", "B6", "<html>Bee-<BR>hive</html>",
			"b2o$o2bo$b2o", "bo$obo$obo$bo"},
		{"ship", "S6", "<html>Ship</html>",
			"2o$obo$b2o", "b2o$obo$2o"},
		{"loaf", "L7", "<html>Loaf</html>",
			"b2o$o2bo$obo$bo", "b2o$o2bo$bobo$2bo",
			"2bo$bobo$o2bo$b2o", "bo$obo$o2bo$b2o" },
		{"boat", "B5", "<html><CENTER>Boat<BR>" +
				"<font size='-1'>(turner)</font></CENTER></html>",
			"2o$obo$bo", "b2o$obo$bo", "bo$obo$b2o", "bo$obo$2o"},
		{"eater", "E7", "<html><CENTER>Eater<BR>" +
				"<font size='-1'>(turner)</font></CENTER></html>",
			"2o$o$b3o$3bo", "2o$obo$2bo$2b2o",
			"2b2o$bobo$bo$2o", "2b2o$3bo$3o$o",
			"o$3o$3bo$2b2o", "2o$bo$bobo$2b2o",
			"2b2o$2bo$obo$2o", "3bo$b3o$o$2o"},
		{"blocks", "2B4", "<html><CENTER>Blocks<BR>" +
				"<font size='-1'>(turner)</font></CENTER></html>",
			"2o$2o4$4b2o$4b2o", "4b2o$4b2o4$2o$2o",
			"5b2o$5b2o3$2o$2o", "2o$2o3$5b2o$5b2o"},
		{"glider", "G", "<html>Glider</html>",
			"3o$o$bo", "bo$2o$obo", "b2o$obo$2bo", "2o$b2o$o",
			"bo$2bo$3o", "obo$b2o$bo", "o$obo$2o", "2bo$2o$b2o"},
		{"paste", "Paste", "<html>Paste</html>", ""}
	};

	private static final int numberOfSeedTypes = seedRLEs.length;
	
	private static final Cell[] gliderStrides =
	{
		new Cell(-1, -1), new Cell(-1, -1),
		new Cell(1, -1), new Cell(1, -1),
		new Cell(1, 1), new Cell(1, 1),
		new Cell(-1, 1), new Cell(-1, 1)
	};

	private static LifePattern[][] seeds = buildSeeds();
	private static String[] seedNames;
	private static String[] seedAbbrevs;
	private static String[] seedLabels;
	
	private static LifePattern[][] buildSeeds()
	{
		LifePattern[][] result = new LifePattern[numberOfSeedTypes][];
		seedNames = new String[numberOfSeedTypes];
		seedAbbrevs = new String[numberOfSeedTypes];
		seedLabels = new String[numberOfSeedTypes];
		
		for (int i = 0; i < numberOfSeedTypes; i++)
		{
			final int PREFIX = 3;
			
			String[] list = seedRLEs[i];
			LifePattern[] patterns = new LifePattern[list.length - PREFIX];
			
			seedNames[i] = list[0];
			seedAbbrevs[i] = list[1];
			seedLabels[i] = list[2];
					
			for (int j = PREFIX; j < list.length; j++)
				patterns[j - PREFIX] =
					Frame.emptyPattern.newPatternFromRLE(list[j]);
			
			result[i] = patterns;
		}

		return result;
	}
	
	public final Nouns nouns = new Nouns();
	
	private int seedTypeIndex = 0;
	private int[] seedOrientIndices = new int[numberOfSeedTypes];

	private PuzzleNode selectedNode;

	private LifePattern placedSeeds;
	private int startGen = 0;
	
	private LifePattern smear;
	private LifePattern forbidden;
	private LifePattern lastseedlessSmooch = Frame.emptyPattern;
	
	private LifePattern startPattern;
	private int gens = INIT_GENS;
	private LifePattern lastEffectiveSeed = Frame.emptyPattern;
	
	private LifePattern[] smears = null;
	private LifePattern debris = null;
	
	private boolean pasteSeedIsStable = true;

	public PuzzleTree()
	{
		super(new PuzzleNode("(root)"));
		
		PuzzleNode thisRoot = getRoot();

		for (int i = Puzzles.puzzles.length - 1; i >= 0; --i)
		{
			PuzzleNode puzzle = Puzzles.puzzles[i];
			String name = puzzle.getName();
			
			selectedNode =
				new PuzzleNode(name, puzzle.getDescription(),
							   puzzle.getPattern(), puzzle.getSmear());
			insertNodeInto(selectedNode, thisRoot, 0);
		}

		selectedNodeChanged();
	}
	
	private void selectedNodeChanged()
	{	
		placedSeeds = Frame.emptyPattern;
		TreeNode[] path = selectedNode.getPath();
		
		for (int i = 2; i < path.length; i++)
		{
			PuzzleNode node = (PuzzleNode) path[i];
			placedSeeds = placedSeeds.or(node.getPattern());
		}
		
		placedSeedsOrStartGenChanged();
	}
	
	private void placedSeedsOrStartGenChanged()
	{
		LifePattern[] ps =
			getOriginalPattern().or(placedSeeds).smear(startGen);
		
		startPattern = ps[0];
		smear = getSelectedPuzzleRoot().getSmear().or(ps[1]);
		forbidden = smear.smooch();
		
		startPatternOrGensOrLastEffectiveSeedChanged();
	}
	
	private void startPatternOrGensOrLastEffectiveSeedChanged()
	{
		LifePattern pattern = getOriginalPattern().or(placedSeeds);
		
		if (effectiveSeedInteractsWithLastSmooch(lastEffectiveSeed))
		{
//			if (seedNames[seedTypeIndex].equals("glider") && startGen > 0)
//			{
//				Cell stride = gliderStrides[seedOrientIndices[seedTypeIndex]];
//				Cell offset = stride.times(startGen / 4);
//				System.out.println(offset);
//				pattern = startPattern.or(lastEffectiveSeed.translate(offset));
//			}
//			else pattern = startPattern.or(lastEffectiveSeed);
			pattern = startPattern.or(lastEffectiveSeed);
		}
		
		smears = new LifePattern[gens / SMEAR_INTERVAL];
		LifePattern last = pattern.getEmpty();
		
		for (int i = 0; i < smears.length; i++)
		{
			LifePattern[] ps = pattern.smear(SMEAR_INTERVAL);
			pattern = ps[0];
			last = smears[i] = ps[1].or(last);
		}

		debris = pattern;
		
		if (lastEffectiveSeed.isEmpty())
			lastseedlessSmooch = smears[smears.length - 1].smooch();
	}

	private boolean currentSeedIsGlider()
	{
		return seedNames[seedTypeIndex].equals("glider");
	}

	private boolean currentSeedIsPaste()
	{
		return seedNames[seedTypeIndex].equals("paste");
	}

	private boolean effectiveSeedInteractsWithLastSmooch
		(LifePattern effectiveSeed)
	{
		return currentSeedIsGlider() || currentSeedIsPaste() ||
			!lastseedlessSmooch.and(effectiveSeed).isEmpty();
	}
	
	public LifePattern[] getSmears(LifePattern effectiveSeed)
	{
		if (effectiveSeed.equals(lastEffectiveSeed))
		{
			lastEffectiveSeed = effectiveSeed;
			return smears;
		}

		if (effectiveSeedInteractsWithLastSmooch(effectiveSeed) ||
			effectiveSeedInteractsWithLastSmooch(lastEffectiveSeed))
		{
			lastEffectiveSeed = effectiveSeed;
			startPatternOrGensOrLastEffectiveSeedChanged();
		}
		else lastEffectiveSeed = effectiveSeed;
		
		return smears;
	}

	public LifePattern getDebris()
	{
		return effectiveSeedInteractsWithLastSmooch(lastEffectiveSeed) ?
			debris : debris.or(lastEffectiveSeed);
	}
	
	public PuzzleNode getSelectedPuzzleRoot()
	{
		return (PuzzleNode) selectedNode.getPath()[1];
	}
	
	public LifePattern getOriginalPattern()
	{
		return getSelectedPuzzleRoot().getPattern();
	}
	
	public String getSelectedSeedAbbrev() {return seedAbbrevs[seedTypeIndex];}
	public String getSelectedSeedLabel() {return seedLabels[seedTypeIndex];}
	public LifePattern getAllPlacedSeeds() {return placedSeeds;}
	public LifePattern getForbidden() {return forbidden;}
	public int getStartGen() {return startGen;}
	public int getGens() {return gens;}
	
	public int getSelectedSeedScore()
	{
		return SCORE_BASE + log2[seeds[seedTypeIndex].length];
	}
	
	@Override
	public PuzzleNode getRoot() {return (PuzzleNode) super.getRoot();}
	
	public LifePattern getSelectedSeed()
	{
		return seeds[seedTypeIndex]
					[seedOrientIndices[seedTypeIndex]];
	}
	
	public LifePattern getOriginalWithSeeds()
	{
		return getOriginalPattern().or(getAllPlacedSeeds());
	}
	
	public LifePattern getLastPlacedSeed()
	{
		return selectedNode.isSeedNode() ? selectedNode.getPattern() : null;
	}

	public TreePath getSelectedPath()
	{
		return new TreePath(selectedNode.getPath());
	}
	
	public String getSelectedPuzzleDescription()
	{
		return getSelectedPuzzleRoot().getDescription();
	}

	public void selectNode(PuzzleNode node)
	{
		if (selectedNode != node)
		{
			selectedNode = node;
			selectedNodeChanged();
		}
	}

	public void selectBlockSeed() {seedTypeIndex = 0;}

	public void incrementSeedType(int i)
	{
		seedTypeIndex = mod(seedTypeIndex + i, numberOfSeedTypes);
		if (getSelectedSeed().isEmpty())
			seedTypeIndex = mod(seedTypeIndex + i, numberOfSeedTypes);
	}
	
	public void incrementSeedOrient(int i)
	{
		seedOrientIndices[seedTypeIndex] =
			mod(seedOrientIndices[seedTypeIndex] + i,
				seeds[seedTypeIndex].length);
	}
	
	public void increaseStartGen(int sign)
	{
		if (sign > 0)
		{
			startGen += START_GEN_INC;
			placedSeedsOrStartGenChanged();
		}
		else if (sign < 0 && startGen > 0)
		{
			startGen -= START_GEN_INC;
			placedSeedsOrStartGenChanged();
		}
	}

	public void resetStartGen()
	{
		if (startGen != 0)
		{
			startGen = 0;
			placedSeedsOrStartGenChanged();
		}
	}
	
	public void resetGens()
	{
		if (gens != INIT_GENS)
		{
			gens = INIT_GENS;
			startPatternOrGensOrLastEffectiveSeedChanged();
		}
	}

	public void increaseGens(int sign)
	{
		if (sign > 0 && gens < MAX_GENS)
		{
			gens = Integer.bitCount(gens) == 1 ?
				gens * 3 / 2 : gens * 4 / 3;
			startPatternOrGensOrLastEffectiveSeedChanged();
		}
		else if (sign < 0 && gens > MIN_GENS)
		{
			gens = Integer.bitCount(gens) == 1 ?
					gens / 4 * 3 : gens / 3 * 2;
			startPatternOrGensOrLastEffectiveSeedChanged();
		}
	}
	
	public void addSelectedSeedAt(Cell mouseCell)
	{
		PuzzleNode parentNode = selectedNode;
		
		int level = parentNode.getLevel();
		
		if (Character.isDigit(parentNode.getName().charAt(0)))
			level = Integer.valueOf(parentNode.getName().split(" ")[0]) + 1;
			
		String abbrev = getSelectedSeedAbbrev();
		String noun = nouns.next();
		
		String name = level + " " + abbrev + " " + noun;
		
		LifePattern currentSeed = getSelectedSeed().translate(mouseCell);
		
		if (currentSeedIsGlider() && startGen > 0)
		{
			Cell stride = gliderStrides[seedOrientIndices[seedTypeIndex]];
			Cell offset = stride.times(startGen / 4).negated();
//			System.out.println(offset);
			currentSeed = currentSeed.translate(offset);
		}

		PuzzleNode node = new PuzzleNode(name, currentSeed);
		int index = parentNode.getChildCount();
		
		this.insertNodeInto(node, parentNode, index);

		selectNode(node);
	}
	
	public final void addPuzzles(PuzzleNode newRoot)
	{
		PuzzleNode thisRoot = getRoot();

		for (int i = 0; i < newRoot.getChildCount(); i++)
		{
			PuzzleNode newPuzzle = (PuzzleNode) newRoot.getChildAt(i);
			
			int n = thisRoot.getChildCount();
			boolean found = false;
			
			for (int j = 0; j < n; j++)
			{
				PuzzleNode oldPuzzle = (PuzzleNode) thisRoot.getChildAt(j);
				
				if (newPuzzle.getName().equals(oldPuzzle.getName()))
				{
					int m = newPuzzle.getChildCount();
					
					for (int k = 0; k < m; k++)
					{
						PuzzleNode newNode =
							(PuzzleNode) newPuzzle.getChildAt(0);
						insertNodeInto
							(newNode, oldPuzzle, oldPuzzle.getChildCount());
					}
					
					found = true;
					break;
				}
			}
			
			if (!found)
				insertNodeInto(newPuzzle, thisRoot, thisRoot.getChildCount());
		}
	}

	public boolean canCopySelectedSubtree()
	{
		if (!selectedNode.isSeedNode()) return false;
		
		String name = selectedNode.getName();
		PuzzleNode thisRoot = getRoot();
		int n = thisRoot.getChildCount();
		
		for (int i = 0; i < n; i++)
		{
			PuzzleNode node = (PuzzleNode) thisRoot.getChildAt(i);
			if (node.getName().equals(name)) return false;
		}
		
		return true;
	}

	public void copySelectedSubtree()
	{
		if (!canCopySelectedSubtree()) return;
		
		PuzzleNode thisRoot = getRoot();
		String name = getSelectedPuzzleRoot().getName();
		String description =
			getSelectedPuzzleDescription().charAt(0) == '(' ?
				getSelectedPuzzleDescription() :
				"(Puzzle derived from " + name + ")";
		
	    PuzzleNode node =
			selectedNode.deepCopyWith(description,
									 getOriginalWithSeeds(), smear);
		insertNodeInto(node, thisRoot, thisRoot.getChildCount());
		selectNode(node);
	}
	
	public String toFormattedLifeHistoryRLE(int cols)
	{
		// green (A): getOriginalWithSeeds()
		// blue (B): getSelectedPuzzleRoot().getSmear().andNot(green)
		// black (.): everything else

		LifePattern pattern = getOriginalWithSeeds();
		LifePattern s = smear.andNot(pattern);
		Box box = pattern.getBoundingBox(). union(s.getBoundingBox());
		return LifePattern.toFormattedRLE(LifePattern.toLifeHistoryRLE(pattern, s),
									  box, "LifeHistory", cols);
	}
	
	public void copySelectedNode(boolean fullCopy)
	{
		String puzzleName = getSelectedPuzzleRoot().getName();
		String puzzleDescription = getSelectedPuzzleDescription();
		String newName = selectedNode.getName();
		String newDescription =
			puzzleDescription.charAt(0) == '(' ?
				puzzleDescription :
				"(Puzzle derived from " + puzzleName + ")";
		
		String comment =
			"#C Seeds of Destruction\n" +
			"#C \"" + newName + "\"\n" +
			"#C " + newDescription + "\n";

		String rle = fullCopy ?
			toFormattedLifeHistoryRLE(60) :
			getOriginalWithSeeds().toFormattedRLE(60);
		
		Clipboard clipboard = Toolkit.getDefaultToolkit().getSystemClipboard();
		clipboard.setContents(new StringSelection(comment + rle), null);
	}

	public void pasteAsNewPuzzle(PuzzleNode node)
	{
		PuzzleNode thisRoot = getRoot();
		insertNodeInto(node, thisRoot, thisRoot.getChildCount());
		selectNode(node);
	}
	
	public void setPasteSeed(PuzzleNode node)
	{
		seedTypeIndex = seeds.length - 1;
		LifePattern seed = node.getPattern().homed();
		seeds[seedTypeIndex][0] = seed;
		pasteSeedIsStable = seed.equals(seed.next());
	}
	
	public LifePattern getCurrentSeedAt(Cell cell)
	{
		return cell != null ? getSelectedSeed().translate(cell) : null;
	}

	public boolean currentSeedIsPlaceableAt(Cell cell)
	{
		if (cell == null) return false;
		
		if (currentSeedIsPaste())
			if (startGen > 0 && !pasteSeedIsStable) return false;
		
		// Q = getCurrentSeedAt(cell)
		// P = getOriginalWithSeeds()
		// S = forbidden
		// S & Q & ~P = E
		// &&
		// Q & ~P != E

		LifePattern p = getCurrentSeedAt(cell).andNot(getOriginalWithSeeds());
		return !p.isEmpty() && forbidden.and(p).isEmpty();
		
//		return forbidden.and(getCurrentSeedAt(cell)).isEmpty();
	}
	
	// Add kbd shortcuts for both seed button mousewheels
	// Cancel seed placement when mouse button up after move from current cell
	// Improve paste seed display
	// Update help
	// Fix copy/paste tooltips
	
	// Done:
	// Add odd glider phases
	// Change comments from # to #C
	// Home pasted seed
	// Fix glider SOD behaviour when start gen > 0
	// Fix pasted SOD behaviour when start gen > 0
	// Reuse help/info window
	// Fix paste seed logic
}	
