package igblan.seedsOfDestruction.ui;

import igblan.life.LifePattern;
import igblan.life.RunPattern;
import igblan.seedsOfDestruction.PuzzleNode;
import igblan.seedsOfDestruction.PuzzleTree;
import java.awt.Color;
import java.awt.EventQueue;
import java.awt.Font;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.net.MalformedURLException;
import java.net.URISyntaxException;
import java.net.URL;
import java.nio.file.Files;
import java.nio.file.Paths;
import java.nio.file.StandardCopyOption;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.swing.UIManager;
import javax.swing.event.TreeModelEvent;
import javax.swing.event.TreeModelListener;
import javax.swing.plaf.FontUIResource;
import javax.swing.tree.DefaultTreeCellRenderer;
import javax.swing.tree.TreePath;

public final class Frame extends javax.swing.JFrame implements TreeModelListener
{
	private DefaultTreeCellRenderer cellRenderer =
			new DefaultTreeCellRenderer();
//	private DefaultTreeCellEditor cellEditor;
	
	public static final LifePattern emptyPattern = RunPattern.empty;
	private PuzzleTree puzzle = new PuzzleTree();
	
	public Frame()
	{
		loadPuzzles();
		cellRenderer.setLeafIcon(null);
		cellRenderer.setOpenIcon(null);
		cellRenderer.setClosedIcon(null);
		UIManager.put("ToolTip.font",
			new FontUIResource("SansSerif", Font.PLAIN, 16));
		UIManager.put("ToolTip.background", new Color(224, 224, 128));
		initComponents();
		expandAllTreeNodes();
//		cellEditor = new DefaultTreeCellEditor(patternTree, cellRenderer);
		puzzle.addTreeModelListener(this);
	}

	/**
	 * This method is called from within the constructor to initialize the form.
	 * WARNING: Do NOT modify this code. The content of this method is always
	 * regenerated by the Form Editor.
	 */
	@SuppressWarnings("unchecked")
    // <editor-fold defaultstate="collapsed" desc="Generated Code">//GEN-BEGIN:initComponents
    private void initComponents()
    {

        javax.swing.JSplitPane jSplitPane1 = new javax.swing.JSplitPane();
        javax.swing.JPanel jPanel1 = new javax.swing.JPanel();
        patternPanel = new PatternPanel(puzzle);
        javax.swing.JPanel jPanel3 = new javax.swing.JPanel();
        javax.swing.JScrollPane jScrollPane1 = new javax.swing.JScrollPane();
        patternTree = new javax.swing.JTree();

        setDefaultCloseOperation(javax.swing.WindowConstants.EXIT_ON_CLOSE);
        setTitle("Seeds Of Destruction");
        setFont(new java.awt.Font("Arial", 0, 14)); // NOI18N
        setPreferredSize(new java.awt.Dimension(800, 640));
        addWindowListener(new java.awt.event.WindowAdapter()
        {
            public void windowClosing(java.awt.event.WindowEvent evt)
            {
                frameClosing(evt);
            }
        });

        jSplitPane1.setDividerLocation(600);
        jSplitPane1.setResizeWeight(1.0);

        jPanel1.setBorder(javax.swing.BorderFactory.createLineBorder(new java.awt.Color(0, 0, 0)));
        jPanel1.setPreferredSize(new java.awt.Dimension(768, 512));
        jPanel1.setLayout(new java.awt.BorderLayout());
        jPanel1.add(patternPanel, java.awt.BorderLayout.CENTER);

        jSplitPane1.setLeftComponent(jPanel1);

        jPanel3.setPreferredSize(new java.awt.Dimension(100, 0));
        jPanel3.setLayout(new java.awt.BorderLayout());

        patternTree.setFont(new java.awt.Font("Tahoma", 1, 14)); // NOI18N
        patternTree.setModel(puzzle);
        patternTree.setCellRenderer(cellRenderer);
        patternTree.setName("patternTree"); // NOI18N
        patternTree.setRootVisible(false);
        patternTree.setSelectionRows(new int[] {0});
        patternTree.setShowsRootHandles(true);
        patternTree.addTreeSelectionListener(new javax.swing.event.TreeSelectionListener()
        {
            public void valueChanged(javax.swing.event.TreeSelectionEvent evt)
            {
                patternTreeSelectionChanged(evt);
            }
        });
        jScrollPane1.setViewportView(patternTree);

        jPanel3.add(jScrollPane1, java.awt.BorderLayout.CENTER);

        jSplitPane1.setRightComponent(jPanel3);

        getContentPane().add(jSplitPane1, java.awt.BorderLayout.CENTER);

        pack();
        setLocationRelativeTo(null);
    }// </editor-fold>//GEN-END:initComponents

    private void patternTreeSelectionChanged(javax.swing.event.TreeSelectionEvent evt)//GEN-FIRST:event_patternTreeSelectionChanged
    {//GEN-HEADEREND:event_patternTreeSelectionChanged
		PuzzleNode node = (PuzzleNode) evt.getPath().getLastPathComponent();
		PuzzleNode puzzleNode = puzzle.getSelectedPuzzleRoot();
		puzzle.selectNode(node);
		
//        expandSelectedSubtree();
		
		if (puzzleNode != puzzle.getSelectedPuzzleRoot())
			((PatternPanel) patternPanel).centerMetricsBox();
		
		patternPanel.repaint();
		
		((PatternPanel) patternPanel).updateArrowButton();
    }//GEN-LAST:event_patternTreeSelectionChanged

    private void frameClosing(java.awt.event.WindowEvent evt)//GEN-FIRST:event_frameClosing
    {//GEN-HEADEREND:event_frameClosing
        savePuzzles();
    }//GEN-LAST:event_frameClosing
	
    // Variables declaration - do not modify//GEN-BEGIN:variables
    private javax.swing.JPanel patternPanel;
    private javax.swing.JTree patternTree;
    // End of variables declaration//GEN-END:variables

	private void expandAllTreeNodes()
	{
		for (int i = 0; i < patternTree.getRowCount(); i++)
			patternTree.expandRow(i);
	}

	public void expandSelectedSubtree()
	{
		TreePath path = puzzle.getSelectedPath();
		int row = patternTree.getRowForPath(path);
		
		for (int i = row; i < patternTree.getRowCount(); i++)
		{
			if (!path.isDescendant(patternTree.getPathForRow(i)))
			{
				patternTree.scrollRowToVisible(i - 1);
				return;
			}
			
			patternTree.expandRow(i);
		}
		
		patternTree.scrollRowToVisible(patternTree.getRowCount() - 1);
	}

	private File getSaveFile(String filename)
	{
		try
		{
			return new File(new URL(getClass().getProtectionDomain()
						.getCodeSource().getLocation(), filename).toURI());
		} catch (MalformedURLException ex)
		{
			Logger.getLogger(PatternPanel.class.getName()).log(Level.SEVERE, null, ex);
		} catch (URISyntaxException ex)
		{
			Logger.getLogger(PatternPanel.class.getName()).log(Level.SEVERE, null, ex);
		}
		
		return null;
	}

    public void savePuzzles()
    {
		try
		{
			File file = getSaveFile("savegame.sod");
			
			if (file.exists())
			{
				File backup = getSaveFile("savegame.bak");
				Files.copy(Paths.get(file.getCanonicalPath()),
						   Paths.get(backup.getCanonicalPath()),
						   StandardCopyOption.REPLACE_EXISTING);
				file.delete();
			}
			
			file.createNewFile();
			
			try (ObjectOutputStream stream =
				new ObjectOutputStream(new FileOutputStream(file)))
			{
				stream.writeObject(puzzle.getRoot());
			}
		} catch (IOException ex)
		{
			Logger.getLogger(PatternPanel.class.getName()).log(Level.SEVERE, null, ex);
		}
    }
	
	private void loadPuzzles()
	{
		try
		{
			File file = getSaveFile("savegame.sod");

			try (ObjectInputStream stream =
				new ObjectInputStream(new FileInputStream(file)))
			{
				puzzle.addPuzzles((PuzzleNode) stream.readObject());
			}
		} catch (IOException ex)
		{
		} catch (ClassNotFoundException ex)
		{
			Logger.getLogger(PatternPanel.class.getName()).log(Level.SEVERE, null, ex);
		}
	}
	
	@Override
	public void treeNodesInserted(TreeModelEvent evt)
	{
		EventQueue.invokeLater(new Runnable()
		{
			@Override
			public void run()
			{
				patternTree.setSelectionPath(puzzle.getSelectedPath());
//				expandSelectedSubtree();
			}
		});
	}

	@Override
	public void treeNodesChanged(TreeModelEvent evt)
	{
		throw new UnsupportedOperationException("Not supported yet."); //To change body of generated methods, choose Tools | Templates.
	}

	@Override
	public void treeNodesRemoved(TreeModelEvent evt)
	{
		throw new UnsupportedOperationException("Not supported yet."); //To change body of generated methods, choose Tools | Templates.
	}

	@Override
	public void treeStructureChanged(TreeModelEvent evt)
	{
		throw new UnsupportedOperationException("Not supported yet."); //To change body of generated methods, choose Tools | Templates.
	}

	/**
	 * @param args the command line arguments
	 */
	public static void main(String args[])
	{
		/* Set the Nimbus look and feel */
		//<editor-fold defaultstate="collapsed" desc=" Look and feel setting code (optional) ">
        /* If Nimbus (introduced in Java SE 6) is not available, stay with the default look and feel.
		 * For details see http://download.oracle.com/javase/tutorial/uiswing/lookandfeel/plaf.html 
		 */
		try
		{
			for (javax.swing.UIManager.LookAndFeelInfo info : javax.swing.UIManager.getInstalledLookAndFeels())
			{
				if ("Metal".equals(info.getName()))
				{
					javax.swing.UIManager.setLookAndFeel(info.getClassName());
					break;
				}
			}
		} catch (ClassNotFoundException | InstantiationException |
				 IllegalAccessException |
				 javax.swing.UnsupportedLookAndFeelException ex)
		{
			java.util.logging.Logger.getLogger(Frame.class.getName()).log(java.util.logging.Level.SEVERE, null, ex);
		}
		//</editor-fold>

		/* Create and display the form */
		EventQueue.invokeLater(new Runnable()
		{
			@Override
			public void run()
			{
				new Frame().setVisible(true);
			}
		});
	}
}
