package igblan.seedsOfDestruction.ui;

import igblan.life.Box;
import igblan.life.Cell;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Point;
import java.awt.Rectangle;

/**
 *
 * @author Paul Chapman
 */
public final class Metrics
{
	private static int mod(int x, int y)
	{
		return (x + Math.max((-x / y + 1) * y , 0)) % y;
	}
	
	private Box box = null;
	private int scale;
	private final boolean vertexCentered;
	private Point origin;
	
	Metrics(Box box, int scale, boolean vertextCentered)
	{
		this.box = box;
		this.scale = scale;
		this.vertexCentered = vertextCentered;
		update(new Dimension(256, 256));
	}
	
	public void update(Dimension size)
	{
		int w = size.width;
		int h = size.height;
		
		int px = vertexCentered ? 0 : (box.left() + box.right()) & 1;
		int py = vertexCentered ? 0 : (box.top() + box.bottom()) & 1;
		
		int x0 = - mod((px * scale - w) / 2, scale);
		int y0 = - mod((py * scale - h) / 2, scale);
		
		int left = (box.left() + box.right() + (2 * x0 - w) / scale ) / 2;
		int top = (box.top() + box.bottom() + (2 * y0 - h) / scale ) / 2;
		int right = left + (w - 2 * x0) / scale;
		int bottom = top + (h - 2 * y0) / scale;
		
		origin = new Point(x0, y0);
		box = new Box(left, top, right, bottom);
	}
	
	public void update(int scale, Dimension size)
	{
		this.scale = scale;
		update(size);
	}
	
	public Box getBox() {return box;}
	public int getScale() {return scale;}

	public Cell cellAt(Point p)
	{
		return new Cell(
			box.left() + (p.x - origin.x) / scale,
			box.top() + (p.y - origin.y) / scale);
	}

	public Rectangle cellRect(int x, int y)
	{
		return new Rectangle(origin.x + (x - box.left()) * scale,
							 origin.y + (y - box.top()) * scale,
							 scale, scale);
	}
	
	protected void paintGrid(Graphics g)
	{
		Color color = new Color(64, 64, 64);
//		Color color = new Color(64, 64, 64, 128);
//		Color hHighlightColor = new Color(128, 128, 128);
//		Color vHighlightColor = new Color(160, 160, 160);
		
		int h = box.height() * scale;
			
		for (int x = box.left(); x < box.right(); x++)
		{
			int x0 = origin.x + (x - box.left()) * scale;
//			g.setColor(mod(x, 5) == 0 ? hHighlightColor : color);
			g.drawLine(x0, origin.y, x0, origin.y + h);
		}
		
		int w = box.width() * scale;
		
		for (int y = box.top(); y < box.bottom(); y++)
		{
			int y0 = origin.y + (y - box.top()) * scale;
//			g.setColor(mod(y, 5) == 0 ? hHighlightColor : color);
			g.drawLine(origin.x, y0, origin.x + w, y0);
		}
	}

	void translateBox(Cell offset) {box = box.translate(offset);}
}
