package igblan.seedsOfDestruction.ui;

import igblan.life.Box;
import igblan.life.Cell;
import igblan.life.LifePattern;
import igblan.life.PatternPainter;
import igblan.seedsOfDestruction.PuzzleNode;
import igblan.seedsOfDestruction.PuzzleTree;
import java.awt.Color;
import java.awt.Component;
import java.awt.Cursor;
import java.awt.Desktop;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Image;
import java.awt.Rectangle;
import java.awt.Toolkit;
import java.awt.datatransfer.Clipboard;
import java.awt.datatransfer.DataFlavor;
import java.awt.datatransfer.UnsupportedFlavorException;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.MouseEvent;
import java.awt.event.WindowEvent;
import java.awt.image.BufferedImage;
import java.io.BufferedReader;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.io.StringReader;
import java.net.URISyntaxException;
import java.util.Scanner;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.regex.Pattern;
import javax.swing.ImageIcon;
import javax.swing.JEditorPane;
import javax.swing.JFrame;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.Timer;
import javax.swing.event.HyperlinkEvent;
import javax.swing.event.HyperlinkListener;

/**
 *
 * @author Paul Chapman
 */
public final class PatternPanel extends JPanel
{
//	protected static final int SMEAR_BANDS = 2;
//	protected static final int SMEAR_INTERVAL = 4;
	
//	protected static final int INIT_GENS = 1024;
	
	private static final int MIN_SCALE = 3;
	
	private static final Color gridColor = Color.DARK_GRAY;
	
//	protected static final Color originalColor = Color.YELLOW;
//	protected static final Color placedSeedColor = Color.GREEN;
//	protected static final Color forbiddenColor = new Color(0, 0, 255, 96);
	private static final Color originalColor = new Color(0, 192, 255);
	private static final Color placedSeedColor = new Color(0, 192, 0);
	private static final Color lastSeedColor = new Color(0, 255, 0);
	private static final Color forbiddenColor = new Color(255, 255, 0, 96);
	private static final Color accessibleDebrisColor =
		new Color(255, 128, 160);
	private static final Color inaccessibleDebrisColor =
		new Color(255, 64, 64);
	private static final Color seedOutlineColor = Color.BLACK;
	private static final Color activeSeedFillColor = Color.GREEN;
	private static final Color suppressedSeedFillColor = Color.GRAY;
	private static final Color forbiddenSeedFillColor = Color.RED;
	
	private static final PatternPainter
		originalPainter = new TargetPainter(originalColor, originalColor);
	private static final PatternPainter
		placedSeedPainter = new TargetPainter(placedSeedColor, placedSeedColor);
	private static final PatternPainter
		lastSeedPainter = new TargetPainter(Color.BLACK, lastSeedColor);
	
	private static final PatternPainter
		forbiddenPainter = new FillPainter(forbiddenColor);
	
	private static final PatternPainter
		accessibleDebrisPainter = new FillPainter(accessibleDebrisColor);
	private static final PatternPainter
		inaccessibleDebrisPainter = new FillPainter(inaccessibleDebrisColor);
	
	private static final PatternPainter
		activeSeedPainter =
			new TargetPainter(seedOutlineColor, activeSeedFillColor);
	private static final PatternPainter
		suppressedSeedPainter =
			new TargetPainter(seedOutlineColor, suppressedSeedFillColor);
	private static final PatternPainter
		forbiddenSeedPainter =
			new TargetPainter(seedOutlineColor, forbiddenSeedFillColor);
	
//	protected static final PatternPainter[]
//		smearPainters = buildSmearPainters();
//
//	private static PatternPainter[] buildSmearPainters()
//	{
//		PatternPainter[] result = new PatternPainter[SMEAR_BANDS];
//		
//		for (int i = 0; i < result.length; i++)
//		{
//			int n = 64 + 128 * (result.length - i) / result.length;
//			result[i] = new SmearPainter(new Color(n, n, n));
//		}
//		
//		return result;
//	}
	
	private static final Cursor dragCursor = new Cursor(Cursor.HAND_CURSOR);
	
	private static final class Params
	{
		Metrics metrics;
		Graphics g;

		public Params(Metrics metrics, Graphics g)
		{
			this.metrics = metrics;
			this.g = g;
		}
	}
	
	private static class SmearPainter implements PatternPainter
	{
		private final Color color;

		public SmearPainter(Color color) {this.color = color;}
		
		@Override
		public void paintCell(Object o, int x, int y)
		{
			Params params = (Params) o;
			Graphics g = params.g;
			Rectangle rect = params.metrics.cellRect(x, y);
			int scale = params.metrics.getScale() >> 1;
			g.setColor(color);
			g.fillOval(rect.x + 1 - scale, rect.y + 1 - scale,
					   rect.width - 1 + 2 * scale, rect.height - 1 + 2 * scale);
		}
	}
	
	private static class SmearPainterNEW implements PatternPainter
	{
		private final Color color;
		private final int radius;

		public SmearPainterNEW(Color color, int radius)
		{
			this.color = color;
			this.radius = radius;
		}
		
		@Override
		public void paintCell(Object o, int x, int y)
		{
			Params params = (Params) o;
			Graphics g = params.g;
			int halfScale = (params.metrics.getScale() >> 1);
			Rectangle rect = params.metrics.cellRect(x, y);
			rect.grow(radius - halfScale, radius - halfScale);
//			int scale = params.metrics.getScale();
//			int halfScale = scale >> 1;
//			int diam = radius + 1;
//			rect.translate(rect.width / 2, rect.height / 2);
//			rect.translate(- diam / 2, - diam / 2);
//			rect.setSize(diam, diam);
			g.setColor(color);
			g.fillOval(rect.x, rect.y, rect.width, rect.height);
		}
	}
	
	private static class FillPainter implements PatternPainter
	{
		private final Color color;

		public FillPainter(Color color) {this.color = color;}
		
		@Override
		public void paintCell(Object o, int x, int y)
		{
			Params params = (Params) o;
			Graphics g = params.g;
			Rectangle rect = params.metrics.cellRect(x, y);
			g.setColor(color);
			g.fillRect(rect.x + 1, rect.y + 1, rect.width - 1, rect.height - 1);
		}
	}
	
	private static class GridPainter implements PatternPainter
	{
		private final Color color;

		public GridPainter(Color color) {this.color = color;}
		
		@Override
		public void paintCell(Object o, int x, int y)
		{
			Params params = (Params) o;
			Graphics g = params.g;
			Rectangle rect = params.metrics.cellRect(x, y);
			g.setColor(color);
			g.drawRect(rect.x, rect.y, rect.width, rect.height);
		}
	}
	
	private static class TargetPainter implements PatternPainter
	{
		private final Color outlineColor;
		private final Color fillColor;

		public TargetPainter(Color outlineColor, Color fillColor)
		{
			this.outlineColor = outlineColor;
			this.fillColor = fillColor;
		}
		
		@Override
		public void paintCell(Object o, int x, int y)
		{
			Params params = (Params) o;
			Graphics g = params.g;
			Rectangle rect = params.metrics.cellRect(x, y);
			int scale = params.metrics.getScale();
			if (scale == 4)
			{
				rect.translate(2, 1);
				rect.setSize(1, 3);
				g.setColor(fillColor);
				g.fillRect(rect.x, rect.y, rect.width, rect.height);
				return;
			}

			Color color = outlineColor;

			if (scale >= 16) rect.grow(- scale / 6 - 1, - scale / 6 - 1);
			else
			{
				color = fillColor;
				rect.grow(-2, -2);
			}
			
			g.setColor(fillColor);
			g.fillOval(rect.x, rect.y, rect.width, rect.height);
			g.setColor(color);
			g.drawOval(rect.x, rect.y, rect.width, rect.height);
		}
	}

	public static final String helpText = buildHelpText();
	
	private static String buildHelpText()
	{
		try
		{
			String result;
			try (InputStream stream =
					PatternPanel.class.getResourceAsStream("html/help.html"))
			{
				result = new Scanner(stream).useDelimiter("\\A").next();
			}
			return result;
		} catch (FileNotFoundException ex)
		{
			Logger.getLogger(PatternPanel.class.getName()).log(Level.SEVERE, null, ex);
		} catch (IOException ex)
		{
			Logger.getLogger(PatternPanel.class.getName()).log(Level.SEVERE, null, ex);
		}
		return null;
	}
	
	static Pattern lifePattern = Pattern.compile(
		"(#C.*\n)*" +
		"[Xx] *= *\\d*, *[Yy] *= *\\d*, *" +
		"[Rr][Uu][Ll][Ee] *= *([Bb]3/[Ss]23|[Ss]23/[Bb]3)\n" +
		"([\\d\\.OoBb$]*\\n)*" +
		"([\\d\\.OoBb$]*!\\n?)*" +
		"(.*\n)*"
		);

	static Pattern lifeHistoryPattern = Pattern.compile(
		"(#C.*\n)*" +
		"[Xx] *= *\\d*, *[Yy] *= *\\d*, *" +
		"[Rr][Uu][Ll][Ee] *= *LifeHistory\n" +
		"([\\d\\.ABD$]*\\n)*" +
		"([\\d\\.ABD$]*!\\n?)*" +
		"(.*\n)*"
		);

	private static JFrame popupFrame = buildPopupFrame();
	
	private static JFrame buildPopupFrame()
	{
		JEditorPane editor = new JEditorPane("text/html", "");
		editor.setEditable(false);
		
		editor.addHyperlinkListener(new HyperlinkListener()
		{
			@Override
			public void hyperlinkUpdate(HyperlinkEvent e)
			{
				if(e.getEventType() == HyperlinkEvent.EventType.ACTIVATED)
					try
					{
						Desktop.getDesktop().browse(e.getURL().toURI());
					} catch (URISyntaxException | IOException ex)
					{
						Logger.getLogger(PatternPanel.class.getName()).log(Level.SEVERE, null, ex);
					}
			}
		});
		
        final JScrollPane scroll = new JScrollPane(editor);
		
		JFrame result = new JFrame();
		result.add(scroll);
		
		result.addWindowListener(new java.awt.event.WindowAdapter()
        {
			@Override
			public void windowOpened(WindowEvent we)
			{
				super.windowOpened(we);
				scroll.getVerticalScrollBar().setValue(0);
			}
        });
		
		return result;
	}
	
	private final Color pasteButtonBackColor;
	private final ImageIcon seedButtonIcon =
		new ImageIcon(new BufferedImage(64, 64, BufferedImage.TYPE_INT_ARGB));
	private final Metrics metrics;
	
	private final ActionListener pasteButtonFlashTimerListener =
		new ActionListener()
	{
		int count = 0;

		@Override
		public void actionPerformed(ActionEvent ae)
		{
			if (count == 4)
			{
				((Timer) ae.getSource()).stop();
				count = 0;
				return;
			}

			count += 1;
			pasteButton.setBackground
				((count & 1) != 0 ? new Color(255, 128, 128) :
									pasteButtonBackColor);
			updatePasteButton();
		}	
	};
		
	private final Timer pasteButtonFlashTimer =
		new Timer(250, pasteButtonFlashTimerListener);

	private PuzzleTree puzzle;

	private Cell mouseCell = null;
	private Cell dragCell = null;
	private Dimension lastSize = null;
	
	private boolean fullCopy = true;
	private boolean fullPaste = true;
	
	public PatternPanel(PuzzleTree puzzle)
	{
		initComponents();
		this.puzzle = puzzle;
		metrics = new Metrics(puzzle.getForbidden().getBoundingBox(), 8, true);
		arrowButton.setVisible(false);
		pasteButtonBackColor = pasteButton.getBackground();
		updateSeedButtons();
		updateStartgenButton();
		updateGensButton();
		updateCopyButton();
		updatePasteButton();
		updateArrowButton();
	}

	@SuppressWarnings("unchecked")
    // <editor-fold defaultstate="collapsed" desc="Generated Code">//GEN-BEGIN:initComponents
    private void initComponents()
    {

        seedpickButton = new javax.swing.JButton();
        startgenButton = new javax.swing.JButton();
        gensButton = new javax.swing.JButton();
        copyButton = new javax.swing.JButton();
        seedorientButton = new javax.swing.JButton();
        helpButton = new javax.swing.JButton();
        infoButton = new javax.swing.JButton();
        arrowButton = new javax.swing.JButton();
        pasteButton = new javax.swing.JButton();

        setBackground(new java.awt.Color(0, 0, 0));
        setAlignmentX(0.9F);
        setFont(new java.awt.Font("Arial", 0, 16)); // NOI18N
        addMouseWheelListener(new java.awt.event.MouseWheelListener()
        {
            public void mouseWheelMoved(java.awt.event.MouseWheelEvent evt)
            {
                PatternPanel.this.mouseWheelMoved(evt);
            }
        });
        addMouseListener(new java.awt.event.MouseAdapter()
        {
            public void mouseEntered(java.awt.event.MouseEvent evt)
            {
                PatternPanel.this.mouseEntered(evt);
            }
            public void mouseExited(java.awt.event.MouseEvent evt)
            {
                PatternPanel.this.mouseExited(evt);
            }
            public void mousePressed(java.awt.event.MouseEvent evt)
            {
                PatternPanel.this.mousePressed(evt);
            }
            public void mouseReleased(java.awt.event.MouseEvent evt)
            {
                mouserReleased(evt);
            }
        });
        addComponentListener(new java.awt.event.ComponentAdapter()
        {
            public void componentResized(java.awt.event.ComponentEvent evt)
            {
                placeButtons(evt);
            }
            public void componentShown(java.awt.event.ComponentEvent evt)
            {
                placeButtons(evt);
            }
        });
        addMouseMotionListener(new java.awt.event.MouseMotionAdapter()
        {
            public void mouseDragged(java.awt.event.MouseEvent evt)
            {
                PatternPanel.this.mouseDragged(evt);
            }
            public void mouseMoved(java.awt.event.MouseEvent evt)
            {
                PatternPanel.this.mouseMoved(evt);
            }
        });
        setLayout(null);

        seedpickButton.setBackground(new java.awt.Color(126, 76, 103));
        seedpickButton.setFont(new java.awt.Font("SansSerif", 1, 18)); // NOI18N
        seedpickButton.setForeground(new java.awt.Color(255, 255, 255));
        seedpickButton.setToolTipText("<html>Mousewheel to pick seed<br>Click to restore block seed</html>");
        seedpickButton.setBorder(javax.swing.BorderFactory.createCompoundBorder(javax.swing.BorderFactory.createLineBorder(new java.awt.Color(0, 0, 0)), javax.swing.BorderFactory.createBevelBorder(javax.swing.border.BevelBorder.RAISED)));
        seedpickButton.setCursor(new java.awt.Cursor(java.awt.Cursor.N_RESIZE_CURSOR));
        seedpickButton.setDefaultCapable(false);
        seedpickButton.setFocusPainted(false);
        seedpickButton.setFocusable(false);
        seedpickButton.setMargin(new java.awt.Insets(0, 0, 0, 0));
        seedpickButton.setMaximumSize(new java.awt.Dimension(64, 64));
        seedpickButton.setMinimumSize(new java.awt.Dimension(64, 64));
        seedpickButton.setName("seedpick"); // NOI18N
        seedpickButton.setPreferredSize(new java.awt.Dimension(64, 64));
        seedpickButton.setRequestFocusEnabled(false);
        seedpickButton.setRolloverEnabled(true);
        seedpickButton.addMouseWheelListener(new java.awt.event.MouseWheelListener()
        {
            public void mouseWheelMoved(java.awt.event.MouseWheelEvent evt)
            {
                seedpickButtonMouseWheelMoved(evt);
            }
        });
        seedpickButton.addActionListener(new java.awt.event.ActionListener()
        {
            public void actionPerformed(java.awt.event.ActionEvent evt)
            {
                seedpickButtonAction(evt);
            }
        });
        add(seedpickButton);
        seedpickButton.setBounds(10, 10, 64, 64);

        startgenButton.setBackground(new java.awt.Color(74, 117, 134));
        startgenButton.setFont(new java.awt.Font("SansSerif", 1, 20)); // NOI18N
        startgenButton.setForeground(new java.awt.Color(255, 255, 255));
        startgenButton.setToolTipText("<html>Mousewheel to change start generation<br>Click to set to 0</html>");
        startgenButton.setAlignmentX(0.9F);
        startgenButton.setBorder(javax.swing.BorderFactory.createCompoundBorder(javax.swing.BorderFactory.createLineBorder(new java.awt.Color(0, 0, 0)), javax.swing.BorderFactory.createBevelBorder(javax.swing.border.BevelBorder.RAISED)));
        startgenButton.setCursor(new java.awt.Cursor(java.awt.Cursor.N_RESIZE_CURSOR));
        startgenButton.setDefaultCapable(false);
        startgenButton.setFocusPainted(false);
        startgenButton.setFocusable(false);
        startgenButton.setMargin(new java.awt.Insets(0, 0, 0, 0));
        startgenButton.setMaximumSize(new java.awt.Dimension(64, 64));
        startgenButton.setMinimumSize(new java.awt.Dimension(64, 64));
        startgenButton.setName("firstgen"); // NOI18N
        startgenButton.setPreferredSize(new java.awt.Dimension(64, 64));
        startgenButton.setRequestFocusEnabled(false);
        startgenButton.setRolloverEnabled(true);
        startgenButton.addMouseWheelListener(new java.awt.event.MouseWheelListener()
        {
            public void mouseWheelMoved(java.awt.event.MouseWheelEvent evt)
            {
                startgenButtonMouseWheelMoved(evt);
            }
        });
        startgenButton.addActionListener(new java.awt.event.ActionListener()
        {
            public void actionPerformed(java.awt.event.ActionEvent evt)
            {
                startgenButtonAction(evt);
            }
        });
        add(startgenButton);
        startgenButton.setBounds(170, 10, 64, 64);

        gensButton.setBackground(new java.awt.Color(74, 117, 134));
        gensButton.setFont(new java.awt.Font("SansSerif", 1, 20)); // NOI18N
        gensButton.setForeground(new java.awt.Color(255, 255, 255));
        gensButton.setToolTipText("<html>Mousewheel to change number of generations<br>Click to set to " + puzzle.INIT_GENS + "</html>");
        gensButton.setAlignmentX(0.9F);
        gensButton.setBorder(javax.swing.BorderFactory.createCompoundBorder(javax.swing.BorderFactory.createLineBorder(new java.awt.Color(0, 0, 0)), javax.swing.BorderFactory.createBevelBorder(javax.swing.border.BevelBorder.RAISED)));
        gensButton.setCursor(new java.awt.Cursor(java.awt.Cursor.N_RESIZE_CURSOR));
        gensButton.setDefaultCapable(false);
        gensButton.setFocusPainted(false);
        gensButton.setFocusable(false);
        gensButton.setMargin(new java.awt.Insets(0, 0, 0, 0));
        gensButton.setMaximumSize(new java.awt.Dimension(64, 64));
        gensButton.setMinimumSize(new java.awt.Dimension(64, 64));
        gensButton.setName("gens"); // NOI18N
        gensButton.setPreferredSize(new java.awt.Dimension(64, 64));
        gensButton.setRequestFocusEnabled(false);
        gensButton.setRolloverEnabled(true);
        gensButton.addMouseWheelListener(new java.awt.event.MouseWheelListener()
        {
            public void mouseWheelMoved(java.awt.event.MouseWheelEvent evt)
            {
                gensButtonMouseWheelMoved(evt);
            }
        });
        gensButton.addActionListener(new java.awt.event.ActionListener()
        {
            public void actionPerformed(java.awt.event.ActionEvent evt)
            {
                gensButtonAction(evt);
            }
        });
        add(gensButton);
        gensButton.setBounds(240, 10, 64, 64);

        copyButton.setBackground(new java.awt.Color(153, 153, 0));
        copyButton.setFont(new java.awt.Font("SansSerif", 1, 14)); // NOI18N
        copyButton.setIcon(new javax.swing.ImageIcon(getClass().getResource("/igblan/seedsOfDestruction/ui/icons/copyicon.png"))); // NOI18N
        copyButton.setText("<html><div style='background-color:#808080'><table cellpadding='0' cellspacing='0'><tr><td>Full</table></div></html>");
        copyButton.setToolTipText("Copy pattern to clipboard");
        copyButton.setAlignmentX(0.9F);
        copyButton.setBorder(javax.swing.BorderFactory.createCompoundBorder(javax.swing.BorderFactory.createLineBorder(new java.awt.Color(0, 0, 0)), javax.swing.BorderFactory.createBevelBorder(javax.swing.border.BevelBorder.RAISED)));
        copyButton.setCursor(new java.awt.Cursor(java.awt.Cursor.N_RESIZE_CURSOR));
        copyButton.setDefaultCapable(false);
        copyButton.setFocusPainted(false);
        copyButton.setFocusable(false);
        copyButton.setHorizontalTextPosition(javax.swing.SwingConstants.CENTER);
        copyButton.setMargin(new java.awt.Insets(0, 0, 0, 0));
        copyButton.setMaximumSize(new java.awt.Dimension(64, 64));
        copyButton.setMinimumSize(new java.awt.Dimension(64, 64));
        copyButton.setName("copy"); // NOI18N
        copyButton.setPreferredSize(new java.awt.Dimension(64, 64));
        copyButton.setRequestFocusEnabled(false);
        copyButton.setRolloverEnabled(true);
        copyButton.addMouseWheelListener(new java.awt.event.MouseWheelListener()
        {
            public void mouseWheelMoved(java.awt.event.MouseWheelEvent evt)
            {
                copyButtonMouseWheelMoved(evt);
            }
        });
        copyButton.addActionListener(new java.awt.event.ActionListener()
        {
            public void actionPerformed(java.awt.event.ActionEvent evt)
            {
                copyButtonAction(evt);
            }
        });
        add(copyButton);
        copyButton.setBounds(10, 80, 64, 64);

        seedorientButton.setBackground(new java.awt.Color(126, 76, 103));
        seedorientButton.setFont(new java.awt.Font("SansSerif", 0, 18)); // NOI18N
        seedorientButton.setForeground(new java.awt.Color(255, 255, 255));
        seedorientButton.setToolTipText("<html>Mousewheel to orient seed<br>Click to restore block seed</html>");
        seedorientButton.setAlignmentX(0.9F);
        seedorientButton.setBorder(javax.swing.BorderFactory.createCompoundBorder(javax.swing.BorderFactory.createLineBorder(new java.awt.Color(0, 0, 0)), javax.swing.BorderFactory.createBevelBorder(javax.swing.border.BevelBorder.RAISED)));
        seedorientButton.setCursor(new java.awt.Cursor(java.awt.Cursor.N_RESIZE_CURSOR));
        seedorientButton.setDefaultCapable(false);
        seedorientButton.setFocusPainted(false);
        seedorientButton.setFocusable(false);
        seedorientButton.setMargin(new java.awt.Insets(0, 0, 0, 0));
        seedorientButton.setMaximumSize(new java.awt.Dimension(64, 64));
        seedorientButton.setMinimumSize(new java.awt.Dimension(64, 64));
        seedorientButton.setName("seedorient"); // NOI18N
        seedorientButton.setPreferredSize(new java.awt.Dimension(64, 64));
        seedorientButton.setRequestFocusEnabled(false);
        seedorientButton.setRolloverEnabled(true);
        seedorientButton.addMouseWheelListener(new java.awt.event.MouseWheelListener()
        {
            public void mouseWheelMoved(java.awt.event.MouseWheelEvent evt)
            {
                seedorientButtonMouseWheelMoved(evt);
            }
        });
        seedorientButton.addActionListener(new java.awt.event.ActionListener()
        {
            public void actionPerformed(java.awt.event.ActionEvent evt)
            {
                seedpickButtonAction(evt);
            }
        });
        add(seedorientButton);
        seedorientButton.setBounds(90, 10, 64, 64);

        helpButton.setBackground(new java.awt.Color(25, 88, 183));
        helpButton.setFont(new java.awt.Font("SansSerif", 1, 20)); // NOI18N
        helpButton.setForeground(new java.awt.Color(255, 255, 255));
        helpButton.setIcon(new javax.swing.ImageIcon(getClass().getResource("/igblan/seedsOfDestruction/ui/icons/helpicon.png"))); // NOI18N
        helpButton.setToolTipText("Show game help");
        helpButton.setAlignmentX(0.9F);
        helpButton.setBorder(javax.swing.BorderFactory.createCompoundBorder(javax.swing.BorderFactory.createLineBorder(new java.awt.Color(0, 0, 0)), javax.swing.BorderFactory.createBevelBorder(javax.swing.border.BevelBorder.RAISED)));
        helpButton.setCursor(new java.awt.Cursor(java.awt.Cursor.DEFAULT_CURSOR));
        helpButton.setDefaultCapable(false);
        helpButton.setFocusPainted(false);
        helpButton.setFocusable(false);
        helpButton.setMargin(new java.awt.Insets(0, 0, 0, 0));
        helpButton.setMaximumSize(new java.awt.Dimension(64, 64));
        helpButton.setMinimumSize(new java.awt.Dimension(64, 64));
        helpButton.setName("help"); // NOI18N
        helpButton.setPreferredSize(new java.awt.Dimension(64, 64));
        helpButton.setRequestFocusEnabled(false);
        helpButton.setRolloverEnabled(true);
        helpButton.addActionListener(new java.awt.event.ActionListener()
        {
            public void actionPerformed(java.awt.event.ActionEvent evt)
            {
                helpButtonAction(evt);
            }
        });
        add(helpButton);
        helpButton.setBounds(240, 80, 64, 64);

        infoButton.setBackground(new java.awt.Color(25, 88, 183));
        infoButton.setFont(new java.awt.Font("SansSerif", 1, 20)); // NOI18N
        infoButton.setForeground(new java.awt.Color(255, 255, 255));
        infoButton.setIcon(new javax.swing.ImageIcon(getClass().getResource("/igblan/seedsOfDestruction/ui/icons/infoicon.png"))); // NOI18N
        infoButton.setToolTipText("Show info about current puzzle");
        infoButton.setAlignmentX(0.9F);
        infoButton.setBorder(javax.swing.BorderFactory.createCompoundBorder(javax.swing.BorderFactory.createLineBorder(new java.awt.Color(0, 0, 0)), javax.swing.BorderFactory.createBevelBorder(javax.swing.border.BevelBorder.RAISED)));
        infoButton.setCursor(new java.awt.Cursor(java.awt.Cursor.DEFAULT_CURSOR));
        infoButton.setDefaultCapable(false);
        infoButton.setFocusPainted(false);
        infoButton.setFocusable(false);
        infoButton.setMargin(new java.awt.Insets(0, 0, 0, 0));
        infoButton.setMaximumSize(new java.awt.Dimension(64, 64));
        infoButton.setMinimumSize(new java.awt.Dimension(64, 64));
        infoButton.setName("info"); // NOI18N
        infoButton.setPreferredSize(new java.awt.Dimension(64, 64));
        infoButton.setRequestFocusEnabled(false);
        infoButton.setRolloverEnabled(true);
        infoButton.addActionListener(new java.awt.event.ActionListener()
        {
            public void actionPerformed(java.awt.event.ActionEvent evt)
            {
                infoButtonAction(evt);
            }
        });
        add(infoButton);
        infoButton.setBounds(170, 80, 64, 64);

        arrowButton.setBackground(new java.awt.Color(124, 148, 80));
        arrowButton.setFont(new java.awt.Font("SansSerif", 1, 20)); // NOI18N
        arrowButton.setForeground(new java.awt.Color(255, 255, 255));
        arrowButton.setIcon(new javax.swing.ImageIcon(getClass().getResource("/igblan/seedsOfDestruction/ui/icons/arrowicon.png"))); // NOI18N
        arrowButton.setToolTipText("Copy subtree as new puzzle");
        arrowButton.setAlignmentX(0.9F);
        arrowButton.setBorder(javax.swing.BorderFactory.createCompoundBorder(javax.swing.BorderFactory.createLineBorder(new java.awt.Color(0, 0, 0)), javax.swing.BorderFactory.createBevelBorder(javax.swing.border.BevelBorder.RAISED)));
        arrowButton.setDefaultCapable(false);
        arrowButton.setFocusPainted(false);
        arrowButton.setFocusable(false);
        arrowButton.setMargin(new java.awt.Insets(0, 0, 0, 0));
        arrowButton.setMaximumSize(new java.awt.Dimension(64, 64));
        arrowButton.setMinimumSize(new java.awt.Dimension(64, 64));
        arrowButton.setName("arrow"); // NOI18N
        arrowButton.setPreferredSize(new java.awt.Dimension(64, 64));
        arrowButton.setRequestFocusEnabled(false);
        arrowButton.setRolloverEnabled(true);
        arrowButton.addActionListener(new java.awt.event.ActionListener()
        {
            public void actionPerformed(java.awt.event.ActionEvent evt)
            {
                arrowButtonAction(evt);
            }
        });
        add(arrowButton);
        arrowButton.setBounds(10, 320, 64, 64);

        pasteButton.setBackground(new java.awt.Color(153, 153, 0));
        pasteButton.setFont(new java.awt.Font("SansSerif", 1, 14)); // NOI18N
        pasteButton.setIcon(new javax.swing.ImageIcon(getClass().getResource("/igblan/seedsOfDestruction/ui/icons/pasteicon.png"))); // NOI18N
        pasteButton.setText("<html><div style='background-color:#808080'><table cellpadding='0' cellspacing='0'><tr><td>Puzzle</table></div></html>");
        pasteButton.setToolTipText("Copy subtree as new puzzle");
        pasteButton.setAlignmentX(0.9F);
        pasteButton.setBorder(javax.swing.BorderFactory.createCompoundBorder(javax.swing.BorderFactory.createLineBorder(new java.awt.Color(0, 0, 0)), javax.swing.BorderFactory.createBevelBorder(javax.swing.border.BevelBorder.RAISED)));
        pasteButton.setCursor(new java.awt.Cursor(java.awt.Cursor.N_RESIZE_CURSOR));
        pasteButton.setDefaultCapable(false);
        pasteButton.setFocusPainted(false);
        pasteButton.setFocusable(false);
        pasteButton.setHorizontalTextPosition(javax.swing.SwingConstants.CENTER);
        pasteButton.setMargin(new java.awt.Insets(0, 0, 0, 0));
        pasteButton.setMaximumSize(new java.awt.Dimension(64, 64));
        pasteButton.setMinimumSize(new java.awt.Dimension(64, 64));
        pasteButton.setName("paste"); // NOI18N
        pasteButton.setPreferredSize(new java.awt.Dimension(64, 64));
        pasteButton.setRequestFocusEnabled(false);
        pasteButton.setRolloverEnabled(true);
        pasteButton.addMouseWheelListener(new java.awt.event.MouseWheelListener()
        {
            public void mouseWheelMoved(java.awt.event.MouseWheelEvent evt)
            {
                pasteButtonMouseWheelMoved(evt);
            }
        });
        pasteButton.addActionListener(new java.awt.event.ActionListener()
        {
            public void actionPerformed(java.awt.event.ActionEvent evt)
            {
                pasteButtonAction(evt);
            }
        });
        add(pasteButton);
        pasteButton.setBounds(90, 80, 64, 64);
    }// </editor-fold>//GEN-END:initComponents

	private Frame getFrame()
	{
		return (Frame) getTopLevelAncestor();
	}

//	private LifePattern getCurrentSeed()
//	{
//		return mouseCell != null ? puzzle.getCurrentSeedAt(mouseCell) : null;
//	}
	
	private boolean currentSeedIsPlaceable()
	{
		return mouseCell != null && puzzle.currentSeedIsPlaceableAt(mouseCell);
	}
	
	protected void paintGrid(Graphics g)
	{
		g.setColor(gridColor);
		metrics.paintGrid(g);
	}

	private void paintPattern(Graphics g, LifePattern p, PatternPainter painter)
	{
		p.paint(painter, new Params(metrics, g), metrics.getBox());
	}

	public void centerMetricsBox()
	{
		Cell oldCenter = metrics.getBox().center();
		Cell newCenter = puzzle.getForbidden().getBoundingBox().center();
		Cell offset = newCenter.minus(oldCenter);
		metrics.translateBox(offset);
	}

	@Override
	protected void paintComponent(Graphics g)
	{
		super.paintComponent(g);
		
		if (dragCell != null)
		{
			Cell offset = dragCell.minus(mouseCell);
			metrics.translateBox(offset);
			mouseCell = dragCell;
		}
		
		if (!getSize().equals(lastSize))
		{
			lastSize = getSize();
			metrics.update(lastSize);
		}
		
		LifePattern seed = puzzle.getCurrentSeedAt(mouseCell);
		LifePattern lastSeed = puzzle.getLastPlacedSeed();
		PatternPainter seedPainter;
		LifePattern effectiveSeed = Frame.emptyPattern;
		
		if (currentSeedIsPlaceable())
		{
			if (dragCell == null)
			{
				effectiveSeed = seed;
				seedPainter = activeSeedPainter;
			}
			else seedPainter = suppressedSeedPainter;
		}
		else seedPainter = forbiddenSeedPainter;

		LifePattern[] smears = puzzle.getSmears(effectiveSeed);

		for (int i = smears.length - 1; i >= 0 ; --i)
		{
			LifePattern p = i > 0 ? smears[i].andNot(smears[i - 1]) : smears[i];
			int n = (i & 1) == 0 ? 255 : 160;
			Color c = new Color(n, n, n);
			int radius = i * metrics.getScale() / smears.length + 1;
			paintPattern(g, p, new SmearPainterNEW(c, radius));
		}

		LifePattern forbidden = puzzle.getForbidden();
		paintPattern(g, forbidden, forbiddenPainter);
		
		LifePattern forbiddenWithSeed = seed != null
			? forbidden.or(seed.smooch())
			: forbidden;

		LifePattern debris = puzzle.getDebris();
		LifePattern accessibleDebris = debris.andNot(forbiddenWithSeed);
		LifePattern inaccessibleDebris = debris.and(forbiddenWithSeed);
		paintPattern(g, accessibleDebris, accessibleDebrisPainter);
		paintPattern(g, inaccessibleDebris, inaccessibleDebrisPainter);
		
		paintPattern(g, puzzle.getOriginalPattern(), originalPainter);
		paintPattern(g, puzzle.getAllPlacedSeeds(), placedSeedPainter);
		
		if (lastSeed != null) paintPattern(g, lastSeed, lastSeedPainter);
		if (seed != null) paintPattern(g, seed, seedPainter);

		paintGrid(g);
	}

    private void mouseMoved(java.awt.event.MouseEvent evt)//GEN-FIRST:event_mouseMoved
    {//GEN-HEADEREND:event_mouseMoved
		Cell p = metrics.cellAt(evt.getPoint());
		
		if (!p.equals(mouseCell))
		{
			mouseCell = p;
			repaint();
		}
    }//GEN-LAST:event_mouseMoved

    private void mouseEntered(java.awt.event.MouseEvent evt)//GEN-FIRST:event_mouseEntered
    {//GEN-HEADEREND:event_mouseEntered
		mouseCell = metrics.cellAt(evt.getPoint());
		repaint();
    }//GEN-LAST:event_mouseEntered

    private void mouseExited(java.awt.event.MouseEvent evt)//GEN-FIRST:event_mouseExited
    {//GEN-HEADEREND:event_mouseExited
		mouseCell = null;
		repaint();
    }//GEN-LAST:event_mouseExited

    private void mousePressed(java.awt.event.MouseEvent evt)//GEN-FIRST:event_mousePressed
    {//GEN-HEADEREND:event_mousePressed
		mouseCell = metrics.cellAt(evt.getPoint());
		
		if (evt.getButton() == MouseEvent.BUTTON1)
		{
			if (currentSeedIsPlaceable())
			{
				puzzle.addSelectedSeedAt(mouseCell);
//				getFrame().expandSelectedSubtree();
			}
		}
		else if (evt.getButton() == MouseEvent.BUTTON3)
		{
			dragCell = mouseCell;
			setCursor(dragCursor);
			repaint();
		}
    }//GEN-LAST:event_mousePressed

    private void mouserReleased(java.awt.event.MouseEvent evt)//GEN-FIRST:event_mouserReleased
    {//GEN-HEADEREND:event_mouserReleased
		mouseCell = metrics.cellAt(evt.getPoint());
        dragCell = null;
		setCursor(new Cursor(Cursor.DEFAULT_CURSOR));
		repaint();
    }//GEN-LAST:event_mouserReleased

    private void mouseDragged(java.awt.event.MouseEvent evt)//GEN-FIRST:event_mouseDragged
    {//GEN-HEADEREND:event_mouseDragged
		Cell p = metrics.cellAt(evt.getPoint());
		
		if (!p.equals(mouseCell))
		{
			mouseCell = p;
			repaint();
		}
    }//GEN-LAST:event_mouseDragged

    private void mouseWheelMoved(java.awt.event.MouseWheelEvent evt)//GEN-FIRST:event_mouseWheelMoved
    {//GEN-HEADEREND:event_mouseWheelMoved
		int scale = metrics.getScale();
		Cell p = metrics.cellAt(evt.getPoint());
        int inc = evt.getWheelRotation();
		
		if (inc > 0 && scale > MIN_SCALE)
			scale = Integer.bitCount(scale) == 1 ?
					scale / 4 * 3 : scale / 3 * 2;
		else if (inc < 0 && scale < 128)
			scale = Integer.bitCount(scale) == 1 ?
					scale * 3 / 2 : scale * 4 / 3;

		metrics.update(scale, lastSize);
		
		Cell q = metrics.cellAt(evt.getPoint());
		metrics.translateBox(p.minus(q));
		repaint();
    }//GEN-LAST:event_mouseWheelMoved

    private void seedpickButtonMouseWheelMoved(java.awt.event.MouseWheelEvent evt)//GEN-FIRST:event_seedpickButtonMouseWheelMoved
    {//GEN-HEADEREND:event_seedpickButtonMouseWheelMoved
        int inc = evt.getWheelRotation();
		
		if (inc > 0) puzzle.incrementSeedType(1);
		else if (inc < 0) puzzle.incrementSeedType(-1);

        updateSeedButtons();
		repaint();
    }//GEN-LAST:event_seedpickButtonMouseWheelMoved

    private void startgenButtonMouseWheelMoved(java.awt.event.MouseWheelEvent evt)//GEN-FIRST:event_startgenButtonMouseWheelMoved
    {//GEN-HEADEREND:event_startgenButtonMouseWheelMoved
		puzzle.increaseStartGen(- evt.getWheelRotation());
		updateStartgenButton();
		repaint();
    }//GEN-LAST:event_startgenButtonMouseWheelMoved

    private void placeButtons(java.awt.event.ComponentEvent evt)//GEN-FIRST:event_placeButtons
    {//GEN-HEADEREND:event_placeButtons
		String[] buttonNames = {"seedpick", "seedorient",
								"firstgen", "gens",
								"copy", "paste",
								"info", "help"};
		Component[] children = getComponents();
		Rectangle bounds = getBounds();
		int x = bounds.x + bounds.width - 76;
		int y = 12;

		for (int i = 0; i < buttonNames.length; i++)
			for (Component c : children)
				if (c.getName().equals(buttonNames[i]))
					c.setLocation(x, y + i * 76);
    }//GEN-LAST:event_placeButtons

    private void copyButtonAction(java.awt.event.ActionEvent evt)//GEN-FIRST:event_copyButtonAction
    {//GEN-HEADEREND:event_copyButtonAction
		puzzle.copySelectedNode(fullCopy);
    }//GEN-LAST:event_copyButtonAction

    private void gensButtonAction(java.awt.event.ActionEvent evt)//GEN-FIRST:event_gensButtonAction
    {//GEN-HEADEREND:event_gensButtonAction
		puzzle.resetGens();
		updateGensButton();
		repaint();
    }//GEN-LAST:event_gensButtonAction

    private void gensButtonMouseWheelMoved(java.awt.event.MouseWheelEvent evt)//GEN-FIRST:event_gensButtonMouseWheelMoved
    {//GEN-HEADEREND:event_gensButtonMouseWheelMoved
		puzzle.increaseGens(- evt.getWheelRotation());
		updateGensButton();
		repaint();
    }//GEN-LAST:event_gensButtonMouseWheelMoved

    private void startgenButtonAction(java.awt.event.ActionEvent evt)//GEN-FIRST:event_startgenButtonAction
    {//GEN-HEADEREND:event_startgenButtonAction
		puzzle.resetStartGen();
		updateStartgenButton();
		repaint();
    }//GEN-LAST:event_startgenButtonAction

    private void seedpickButtonAction(java.awt.event.ActionEvent evt)//GEN-FIRST:event_seedpickButtonAction
    {//GEN-HEADEREND:event_seedpickButtonAction
		puzzle.selectBlockSeed();
        updateSeedButtons();
		repaint();
    }//GEN-LAST:event_seedpickButtonAction

	private void htmlPopup(String title, String html, int w, int h)
	{
		JScrollPane scrollPane =
			(JScrollPane) popupFrame.getContentPane().getComponent(0);
		JEditorPane editor = (JEditorPane) scrollPane.getViewport().getView();
		
		editor.setText(html);
		
		Rectangle bounds = getFrame().getBounds();
		bounds.translate(bounds.width / 2, bounds.height / 2);
		bounds.setSize(0, 0);
		bounds.grow(w / 2, h / 2);
		
		popupFrame.setBounds(bounds);
		popupFrame.setTitle(title);
		popupFrame.setVisible(true);
	}
	
    private void helpButtonAction(java.awt.event.ActionEvent evt)//GEN-FIRST:event_helpButtonAction
    {//GEN-HEADEREND:event_helpButtonAction
		htmlPopup("Seeds of Destruction - Help", helpText, 600, 400);
    }//GEN-LAST:event_helpButtonAction

    private void infoButtonAction(java.awt.event.ActionEvent evt)//GEN-FIRST:event_infoButtonAction
    {//GEN-HEADEREND:event_infoButtonAction
		PuzzleNode node = puzzle.getSelectedPuzzleRoot();
		
		String html =
			"<html>" +
			"<body style='font-size: 16; font-family: sans-serif'>" +
			"<center><h2>" +
			"<u>" + node.getName() + "</u><br>" +
			"</h2></center>" +
			"<center><table width='90%'><tr align='center'><td>" +
			node.getDescription() +
			"</table></center>" +
			"</html>";
		htmlPopup(node.getName() + " - Info", html, 600, 400);
    }//GEN-LAST:event_infoButtonAction

    private void arrowButtonAction(java.awt.event.ActionEvent evt)//GEN-FIRST:event_arrowButtonAction
    {//GEN-HEADEREND:event_arrowButtonAction
		puzzle.copySelectedSubtree();
		getFrame().expandSelectedSubtree();
    }//GEN-LAST:event_arrowButtonAction

    private void seedorientButtonMouseWheelMoved(java.awt.event.MouseWheelEvent evt)//GEN-FIRST:event_seedorientButtonMouseWheelMoved
    {//GEN-HEADEREND:event_seedorientButtonMouseWheelMoved
        int inc = evt.getWheelRotation();
		
		if (inc > 0) puzzle.incrementSeedOrient(1);
		else if (inc < 0) puzzle.incrementSeedOrient(-1);

        updateSeedButtons();
		repaint();
    }//GEN-LAST:event_seedorientButtonMouseWheelMoved

    private void copyButtonMouseWheelMoved(java.awt.event.MouseWheelEvent evt)//GEN-FIRST:event_copyButtonMouseWheelMoved
    {//GEN-HEADEREND:event_copyButtonMouseWheelMoved
        fullCopy = !fullCopy;
		updateCopyButton();
    }//GEN-LAST:event_copyButtonMouseWheelMoved
	private String getComment(String line)
	{
		return line.startsWith("#C") ? line.substring(2).trim() : null;
	}
	
	private PuzzleNode parseRLE(String contents, int matchCode)
	{
		try
		{
			BufferedReader reader =
				new BufferedReader(new StringReader(contents));
			
			String line = reader.readLine();
			String comment = getComment(line);
			
			if (comment != null &&
				comment.toLowerCase().equals("seeds of destruction"))
			{
				line = reader.readLine();
				comment = getComment(line);
			}
			
			String name;
			
			if (comment != null &&
				comment.startsWith("\"") && comment.endsWith("\""))
			{
				name = comment.substring(1, comment.length() - 1).trim();
				line = reader.readLine();
				comment = getComment(line);
			}
			else name = "Puzzle " + puzzle.nouns.next();

			String description = "";
			
			while (comment != null)
			{
				description += comment + " ";
				line = reader.readLine();
				comment = getComment(line);
			}
			
			line = reader.readLine();
			
			String rle = "";

			while (!line.contains("!"))
			{
				rle += line;
				line = reader.readLine();
			}
			
			rle += line;
			
			if (matchCode == 2)
			{
				return new PuzzleNode(name, description,
					Frame.emptyPattern.newPatternFromRLE(rle),
					Frame.emptyPattern);
			}
			else
			{
				return new PuzzleNode(name, description, rle);
			}
			
		} catch (IOException ex)
		{
			Logger.getLogger(PuzzleTree.class.getName()).
				log(Level.SEVERE, null, ex);
		}
		
		return null;
	}
	

    private void pasteButtonAction(java.awt.event.ActionEvent evt)//GEN-FIRST:event_pasteButtonAction
    {//GEN-HEADEREND:event_pasteButtonAction
   		Clipboard clipboard = Toolkit.getDefaultToolkit().getSystemClipboard();
		String contents = null;
		
		try
		{
			contents = (String) clipboard.getData(DataFlavor.stringFlavor);
		} catch (UnsupportedFlavorException | IOException ex)
		{
			Logger.getLogger(PatternPanel.class.getName()).log(Level.SEVERE, null, ex);
		}

		int matchCode =
			lifePattern.matcher(contents).matches() ? 2 :
			lifeHistoryPattern.matcher(contents).matches() ? 4 : 0;
		
		if (matchCode == 0)
		{
			flashPasteButton();
			return;
		}
		
		PuzzleNode node = parseRLE(contents, matchCode);
		
		if (fullPaste)
		{
			puzzle.pasteAsNewPuzzle(node);
			getFrame().expandSelectedSubtree();
		}
		else
		{
			puzzle.setPasteSeed(node);
			updateSeedButtons();
		}
    }//GEN-LAST:event_pasteButtonAction

    private void pasteButtonMouseWheelMoved(java.awt.event.MouseWheelEvent evt)//GEN-FIRST:event_pasteButtonMouseWheelMoved
    {//GEN-HEADEREND:event_pasteButtonMouseWheelMoved
        fullPaste = !fullPaste;
		updatePasteButton();
    }//GEN-LAST:event_pasteButtonMouseWheelMoved

	private void updateSeedButtons()
	{
		String label = puzzle.getSelectedSeedLabel();
		seedpickButton.setText(label);
		
		LifePattern seed = puzzle.getSelectedSeed();
		Box box = seed.getBoundingBox();
		int seedScale = box.width() > 6 ? 4 : box.width() > 4 ? 6 : 12;
		Metrics m = new Metrics(box, seedScale, false);
		m.update(new Dimension(64, 64));

		Image image = seedButtonIcon.getImage();
		Graphics g = image.getGraphics();
		
		PatternPainter p = new FillPainter(Color.GREEN);
		g.setColor(seedpickButton.getBackground());
		g.fillRect(0, 0, 64, 64);
		seed.paint(p, new Params(m, g), m.getBox());

//		String score = String.valueOf(puzzle.getSelectedSeedScore());
//		g.setFont(seedorientButton.getFont());
//		g.setColor(new Color(0, 0, 0, 128));
//		g.fillRect(48, 43, 12, 17);
//		g.setColor(Color.WHITE);
//		g.drawString(score, 49, 58);
		
		seedorientButton.setIcon(seedButtonIcon);
	}
	
	private void updateStartgenButton()
	{
		startgenButton.setText(
			"<html><center>Start<br><font color=\"#00FF00\">" +
			puzzle.getStartGen() +
			"</font></center></html>");
	}

	private void updateGensButton()
	{
		gensButton.setText(
			"<html><center>Gens<br><font color=\"#00FF00\">" +
			puzzle.getGens() +
			"</font></center></html>");
	}

	public void updateCopyButton()
	{
		copyButton.setText(
			"<html><div style='background-color:#E0E0E0'>" +
			"<table cellpadding='0' cellspacing='0'><tr><td>" +
			(fullCopy ? "Golly" : "Life") +
			"</table></div></html>");
	}
	
	public void updatePasteButton()
	{
		pasteButton.setText(
			"<html><div style='background-color:#E0E0E0'>" +
			"<table cellpadding='0' cellspacing='0'><tr><td>" +
			(fullPaste ? "Puzzle" : "Seed") +
			"</table></div></html>");
	}
	
	public final void updateArrowButton()
	{
		arrowButton.setEnabled(puzzle.canCopySelectedSubtree());
	}
	
	private void flashPasteButton()
	{
		if (pasteButtonFlashTimer.isRunning()) return;
		
		pasteButtonFlashTimer.setInitialDelay(0);
		pasteButtonFlashTimer.start();
	}

    // Variables declaration - do not modify//GEN-BEGIN:variables
    private javax.swing.JButton arrowButton;
    private javax.swing.JButton copyButton;
    private javax.swing.JButton gensButton;
    private javax.swing.JButton helpButton;
    private javax.swing.JButton infoButton;
    private javax.swing.JButton pasteButton;
    private javax.swing.JButton seedorientButton;
    private javax.swing.JButton seedpickButton;
    private javax.swing.JButton startgenButton;
    // End of variables declaration//GEN-END:variables
}
